#!/usr/bin/perl -w
# If your copy of perl is not in /usr/bin, please adjust the line above.
#
# Copyright (C) 1998, 1999 VMware, Inc.  All Rights Reserved.
#
# Host configurator for VMware

use strict;

# Constants
my $cEtcDir = '/etc';
my $cInstallDBDir = $cEtcDir . '/vmware';
my $cInstallDBFileName = $cInstallDBDir . '/locations';
my $cConfFlag = $cInstallDBDir . '/not_configured';
my $cKernelModuleDir = '/lib/modules';

# Global variables
my %gSystem;
my %gHelper;
my @gAvailEthIf;
# By convention, vmnet1 is the virtual ethernet interface connected to the
# private virtual network
my $gHostOnlyEthIf = 'vmnet1';

# BEGINNING OF THE SECOND LIBRARY FUNCTIONS
# Global variables
my %gDBAnswer;
my %gDBFile;
my %gDBDir;

# Load the installer database
sub db_load { 
  open(INSTALLDB, '<' . $cInstallDBFileName) or error('Unable to open the installer database ' . $cInstallDBFileName . ' in read-mode.' . "\n\n");
  while (<INSTALLDB>) {
    if (/^answer (\S+) (\S+)/) {
      $gDBAnswer{$1} = $2;
    } elsif (/^answer (\S+)/) {
      $gDBAnswer{$1} = '';
    } elsif (/^remove_answer (\S+)/) {
      delete $gDBAnswer{$1};
    } elsif (/^file (\S+) (\S+)/) {
      $gDBFile{$1} = $2;
    } elsif (/^file (\S+)/) {
      $gDBFile{$1} = 0;
    } elsif (/^remove_file (\S+)/) {
      delete $gDBFile{$1};
    } elsif (/^directory (\S+)/) {
      $gDBDir{$1} = '';
    }
  }
  close(INSTALLDB);
}

# Open the database on disk in append mode
sub db_append {
  if (not open(INSTALLDB, '>>' . $cInstallDBFileName)) {
    error('Unable to open the installer database ' . $cInstallDBFileName . ' in append-mode.' . "\n\n");
  }
  # Force a flush after every write operation.
  # See 'Programming Perl', p. 110
  select((select(INSTALLDB), $| = 1)[0]);
}

# Add a file to the tar installer database
sub db_add_file {
  my $file = shift;
  my $flags = shift;

  if ($flags & 0x1) {
    my @statbuf;

    @statbuf = stat($file);
    if (not (defined($statbuf[9]))) {
      error('Unable to get the last modification timestamp of the destination file ' . $file . '.' . "\n\n");
    }

    $gDBFile{$file} = $statbuf[9];
    print INSTALLDB 'file ' . $file . ' ' . $statbuf[9] . "\n";
  } else {
    $gDBFile{$file} = 0;
    print INSTALLDB 'file ' . $file . "\n";
  }
}

# Remove a file from the tar installer database
sub db_remove_file {
  my $file = shift;

  print INSTALLDB 'remove_file ' . $file . "\n";
  delete $gDBFile{$file};
}

# Determine if a file belongs to the tar installer database
sub db_file_in {
  my $file = shift;

  return defined($gDBFile{$file});
}

# Return the timestamp of an installed file
sub db_file_ts {
  my $file = shift;

  return $gDBFile{$file};
}

# Add a directory to the tar installer database
sub db_add_dir {
  my $dir = shift;

  $gDBDir{$dir} = '';
  print INSTALLDB 'directory ' . $dir . "\n";
}

# Remove an answer from the tar installer database
sub db_remove_answer {
  my $id = shift;

  if (defined($gDBAnswer{$id})) {
    print INSTALLDB 'remove_answer ' . $id . "\n";
    delete $gDBAnswer{$id};
  }
}

# Add an answer to the tar installer database
sub db_add_answer {
  my $id = shift;
  my $value = shift;

  db_remove_answer($id);
  $gDBAnswer{$id} = $value;
  print INSTALLDB 'answer ' . $id . ' ' . $value . "\n";
}

# Retrieve an answer that must be present in the database
sub db_get_answer {
  my $id = shift;

  if (not defined($gDBAnswer{$id})) {
    error('Unable to find the answer ' . $id . ' in the installer database (' . $cInstallDBFileName . '). You may want to re-install VMware.' . "\n\n");
  }

  return $gDBAnswer{$id};
}

# Save the tar installer database
sub db_save {
  close(INSTALLDB);
}
# END OF THE SECOND LIBRARY FUNCTIONS

# BEGINNING OF THE LIBRARY FUNCTIONS
# Constants
my $cTerminalLineSize = 80;

# Global variables
my %gOption;
my %gAnswerSize;
my %gCheckAnswerFct;

# Tell if the user is the super user
sub is_root {
  return $> == 0;
}

# Wordwrap system: append some content to the output
sub append_output {
  my $output = shift;
  my $pos = shift;
  my $append = shift;

  $output .= $append;
  $pos = ($pos + length($append)) % $cTerminalLineSize;

  return ($output, $pos);
}

# Wordwrap system: deal with the next character
sub wrap_one_char {
  my $output = shift;
  my $pos = shift;
  my $word = shift;
  my $char = shift;
  my $reserved = shift;
  my $length;

  if (not (($char eq "\n") || ($char eq ' ') || ($char eq ''))) {
    $word .= $char;

    return ($output, $pos, $word);
  }

  # We found a separator. Process the last word

  $length = length($word) + $reserved;
  if ((($pos + $length) > $cTerminalLineSize) && ($length <= $cTerminalLineSize)) {
    # The word is too large to fit in the end of line, but can fit in a new
    # empty line, let's create one for it
    $output .= "\n";
    $pos = 0;
  }

  ($output, $pos) = append_output($output, $pos, $word);
  $word = '';

  if ($char eq "\n") {
    $output .= "\n";
    $pos = 0;
  } elsif ($char eq ' ') {
    if ($pos) {
      ($output, $pos) = append_output($output, $pos, ' ');
    }
  }

  return ($output, $pos, $word);
}

# Wordwrap system: word-wrap a string plus some reserved trailing space
sub wrap {
  my $input = shift;
  my $reserved = shift;
  my $output;
  my $pos;
  my $word;
  my $i;

  $output = '';
  $pos = 0;
  $word = '';
  for ($i = 0; $i < length($input); $i++) {
    ($output, $pos, $word) = wrap_one_char($output, $pos, $word, substr($input, $i, 1), 0);
  }
  # Use an artifical last '' separator to process the last word
  ($output, $pos, $word) = wrap_one_char($output, $pos, $word, '', $reserved);

  return $output;
}

# Print an error message and exit
sub error {
  my $msg = shift;

  print STDERR wrap($msg . 'Execution aborted.' . "\n\n", 0);
  exit 1;
}

# Convert a string to its equivalent shell representation
sub shell_string {
  my $single_quoted = shift;

  $single_quoted =~ s/'/'"'"'/g;
  # This comment is a fix for emacs's broken syntax-highlighting code --hpreg
  return '\'' . $single_quoted . '\'';
}

# Contrary to a popular belief, 'which' is not always a shell builtin command.
# So we can not trust it to determine the location of other binaries.
# Moreover, SuSE 6.1's 'which' is unable to handle program names beginning with
# a '/'...
#
# Return value is the complete path if found, or '' if not found
sub internal_which {
  my $bin = shift;

  if (substr($bin, 0, 1) eq '/') {
    # Absolute name
    if ((-f $bin) && (-x $bin)) {
      return $bin;
    }
  } else {
    # Relative name
    my @paths;
    my $path;

    if (index($bin, '/') == -1) {
      # There is no other '/' in the name
      @paths = split(':', $ENV{'PATH'});
      foreach $path (@paths) {
	my $fullbin;

	$fullbin = $path . '/' . $bin;
	if ((-f $fullbin) && (-x $fullbin)) {
	  return $fullbin;
	}
      }
    }
  }

  return '';
}

# Remove leading and trailing whitespaces
sub remove_whitespaces {
  my $string = shift;

  $string =~ s/^\s*//;
  $string =~ s/\s*$//;
  return $string;
}

# Ask a question to the user and propose an optional default value
# Use this when you don't care about the validity of the answer
sub query {
    my $message = shift;
    my $defaultreply = shift;
    my $reserved = shift;
    my $reply;

    # Reserve some room for the reply
    print wrap($message . (($defaultreply eq '') ? '' : (' [' . $defaultreply . ']')), 1 + $reserved);
    # This is what the 1 is for
    print ' ';
	   
    if ($gOption{'default'} == 1) {
      # Simulate the enter key
      print "\n";
      $reply = '';
    } else {
      chop($reply = <STDIN>);
    }

    print "\n";
    $reply = remove_whitespaces($reply);
    if ($reply eq '') {
      $reply = $defaultreply;
    }
    return $reply;
}

# Check the validity of an answer whose type is yesno
# Return a clean answer if valid, or ''
sub check_answer_binpath {
  my $answer = shift;
  my $source = shift;

  if (not (internal_which($answer) eq '')) {
    return $answer;
  }

  if ($source eq 'user') {
    print wrap('The answer "' . $answer . '" is invalid. It must be the complete name of a binary file.' . "\n\n", 0);
  }
  return '';
}
$gAnswerSize{'binpath'} = 20;
$gCheckAnswerFct{'binpath'} = \&check_answer_binpath;

# Prompts the user if a binary is not found
# Return value is:
#  '': the binary has not been found
#  the binary name if it has been found
sub DoesBinaryExist_Prompt {
  my $bin = shift;
  my $answer;

  $answer = check_answer_binpath($bin, 'default');
  if (not ($answer eq '')) {
    return $answer;
  }

  if (get_answer('Setup is unable to find the "' . $bin . '" program on your machine. Please make sure it is installed. Do you want to specify the location of this program by hand?', 'yesno', 'yes') eq 'no') {
    return '';
  }

  return get_answer('What is the location of the "' . $bin . '" program on your machine?', 'binpath', '');
}

# Execute the command passed as an argument
# _without_ interpolating variables (Perl does it by default)
sub direct_command {
  return `$_[0]`;
}

# chmod() that reports errors
sub safe_chmod {
  my $mode = shift;
  my $file = shift;

  if (chmod($mode, $file) != 1) {
    error('Unable to change the access rights of the file ' . $file . '.' . "\n\n");
  }
}

# Emulate a simplified ls program for directories
sub internal_ls {
  my $dir = shift;
  my @fn;

  opendir(LS, $dir);
  @fn = grep(!/^\.\.?$/, readdir(LS));
  closedir(LS);

  return @fn;
}

# Install a file permission
sub install_permission {
  my $src = shift;
  my $dst = shift;
  my @statbuf;

  @statbuf = stat($src);
  if (not (defined($statbuf[2]))) {
    error('Unable to get the access rights of the source file ' . $src . '.' . "\n\n");
  }
  safe_chmod($statbuf[2] & 0xFFF, $dst);
}

# Emulate a simplified sed program
# Return 1 if success, 0 if failure
sub internal_sed {
  my $src = shift;
  my $dst = shift;
  my $append = shift;
  my $patchRef = shift;
  my @patchKeys;

  if (not open(SRC, '<' . $src)) {
    return 0;
  }
  if (not open(DST, (($append == 1) ? '>>' : '>') . $dst)) {
    return 0;
  }

  @patchKeys = keys(%$patchRef);
  if ($#patchKeys == -1) {
    while(defined($_ = <SRC>)) {
      print DST $_;
    }
  } else {
    while(defined($_ = <SRC>)) {
      my $patchKey;

      foreach $patchKey (@patchKeys) {
	s/$patchKey/$$patchRef{$patchKey}/g;
      }

      print DST $_;
    }
  }

  close(SRC);
  close(DST);
  return 1;
}

# Check if a file name exists
sub file_name_exist {
  my $file = shift;

  # Note: We must test for -l before, because if an existing symlink points to
  #       a non-existing file, -e will be false
  return ((-l $file) || (-e $file))
}

# Check if a file already exists and prompt the user
# Return 0 if the file can be written safely, 1 otherwise
sub file_check_exist {
  my $file = shift;

  if (not file_name_exist($file)) {
    return 0;
  }

  # The default must make sure that the product will be correctly installed
  # We give the user the choice so that a sysadmin can perform a normal
  # install on a NFS server and then answer 'no' NFS clients
  return (get_answer('The file ' . $file . ' that this script was about to install already exists. Overwrite?', 'yesno', 'yes') eq 'yes') ? 0 : 1;
}

# Install one file
sub install_file {
  my $src = shift;
  my $dst = shift;
  my $patchRef = shift;
  my $ts = shift;

  uninstall_file($dst);
  if (file_check_exist($dst)) {
    return;
  }
  # The file could be a symlink to another location. Remove it
  unlink($dst);
  if (not internal_sed($src, $dst, 0, $patchRef)) {
    error('Unable to copy the source file ' . $src . ' to the destination file ' . $dst . '.' . "\n\n");
  }
  db_add_file($dst, $ts);
  install_permission($src, $dst);
}

# mkdir() that reports errors
sub safe_mkdir {
  my $file = shift;

  if (mkdir($file, 0000) == 0) {
    error('Unable to create the directory ' . $file . '.' . "\n\n");
  }
}

# Remove trailing slashes in a dir path
sub dir_remove_trailing_slashes {
  my $path = shift;

  for(;;) {
    my $len;
    my $pos;

    $len = length($path);
    if ($len < 2) {
      # Could be '/' or any other character. Ok.
      return $path;
    }

    $pos = rindex($path, '/');
    if ($pos != $len - 1) {
      # No trailing slash
      return $path;
    }

    # Remove the trailing slash
    $path = substr($path, 0, $len - 1)
  }
}

# Emulate a simplified dirname program
sub internal_dirname {
  my $path = shift;
  my $pos;

  $path = dir_remove_trailing_slashes($path);

  $pos = rindex($path, '/');
  if ($pos == -1) {
    # No slash
    return '.';
  }

  if ($pos == 0) {
    # The only slash is at the beginning
    return '/';
  }

  return substr($path, 0, $pos);
}

# Create a hierarchy of directories with permission 0755
# Return 1 if the directory existed before
sub create_dir {
  my $dir = shift;
  my $use_db = shift;

  if (-d $dir) {
    return 1;
  }

  if (index($dir, '/') != -1) {
    create_dir(internal_dirname($dir), $use_db);
  }
  safe_mkdir($dir);
  if ($use_db) {
    db_add_dir($dir);
  }
  safe_chmod(0755, $dir);
  return 0;
}

# Get a valid non-persistent answer to a question
# Use this when the answer shouldn't be stored in the database
sub get_answer {
  my $msg = shift;
  my $type = shift;
  my $default = shift;
  my $answer;

  if (not defined($gAnswerSize{$type})) {
    die 'get_answer(): type ' . $type . ' not implemented :(' . "\n\n";
  }
  for (;;) {
    $answer = check_answer(query($msg, $default, $gAnswerSize{$type}), $type, 'user');
    if (not ($answer eq '')) {
      return $answer;
    }
  }
}

# Get a valid persistent answer to a question
# Use this when you want an answer to be stored in the database
sub get_persistent_answer {
  my $msg = shift;
  my $id = shift;
  my $type = shift;
  my $default = shift;
  my $answer;

  if (defined($gDBAnswer{$id})) {
    # There is a previous answer in the database
    $answer = check_answer($gDBAnswer{$id}, $type, 'db');
    if (not ($answer eq '')) {
      # The previous answer is valid. Make it the default value
      $default = $answer;
    }
  }

  $answer = get_answer($msg, $type, $default);
  db_add_answer($id, $answer);
  return $answer;
}

# Find a suitable backup name and backup a file
sub backup_file {
  my $file = shift;
  my $i;

  for ($i = 0; $i < 100; $i++) {
    if (! -e $file . '.old.' . $i) {
      my %patch;

      undef %patch;
      if (internal_sed($file, $file . '.old.' . $i, 0, \%patch)) {
	print wrap('File ' . $file . ' is backed up to ' . $file . '.old.' . $i . '.' . "\n\n", 0);
      } else {
	print STDERR wrap('Unable to backup the file ' . $file . ' to ' . $file . '.old.' . $i .'.' . "\n\n", 0);
      }
      return;
    }
  }

  print STDERR wrap('Unable to backup the file ' . $file . '. You have too many backups files. They are files of the form ' . $file . '.old.N, where N is a number. Please delete some of them.' . "\n\n", 0);
}

# Uninstall a file previously installed by us
sub uninstall_file {
  my $file = shift;

  if (not db_file_in($file)) {
    # Not installed by this script
    return;
  }

  if (file_name_exist($file)) {
    if (db_file_ts($file)) {
      my @statbuf;

      @statbuf = stat($file);
      if (defined($statbuf[9])) {
	if (db_file_ts($file) != $statbuf[9]) {
	  # Modified since this script installed it
	  backup_file($file);
	}
      } else {
	print STDERR wrap('Unable to get the last modification timestamp of the file ' . $file . '.' . "\n\n", 0);
      }
    }
  } else {
    print wrap('This script previously created the file ' . $file . ', and was about to remove it. Somebody else apparently did it already.' . "\n\n", 0);
  }

  if (not unlink($file)) {
    print STDERR wrap('Unable to remove the file ' . $file . '.' . "\n\n", 0);
  }
  db_remove_file($file);
}

# Return the version of VMware
sub vmware_version {
  my $buildNr;

  $buildNr = '1.1.2 ' . q$Name: build-364 $;
  $buildNr =~ s/Name: //;
  return remove_whitespaces($buildNr);
}

# Check the validity of an answer whose type is yesno
# Return a clean answer if valid, or ''
sub check_answer_yesno {
  my $answer = shift;
  my $source = shift;

  if (lc($answer) =~ /^y(es)?$/) {
    return 'yes';
  }

  if (lc($answer) =~ /^n(o)?$/) {
    return 'no';
  }

  if ($source eq 'user') {
    print wrap('The answer "' . $answer . '" is invalid. It must be one of "y" or "n".' . "\n\n", 0);
  }
  return '';
}
$gAnswerSize{'yesno'} = 3;
$gCheckAnswerFct{'yesno'} = \&check_answer_yesno;

# Check the validity of an answer based on its type
# Return a clean answer if valid, or ''
sub check_answer {
  my $answer = shift;
  my $type = shift;
  my $source = shift;

  if (not defined($gCheckAnswerFct{$type})) {
    die 'check_answer(): type ' . $type . ' not implemented :(' . "\n\n";
  }
  return &{$gCheckAnswerFct{$type}}($answer, $source);
}
# END OF THE LIBRARY FUNCTIONS

# Set up the location of external helpers
sub initialize_external_helpers {
  my $program;

  $gHelper{'more'} = '';
  if (defined($ENV{'PAGER'})) {
    my @tokens;

    # The environment variable sometimes contains the pager name _followed by
    # a few command line options_.
    #
    # Isolate the program name (we are certain it does not contain a
    # whitespace) before dealing with it.
    @tokens = split(' ', $ENV{'PAGER'});
    $tokens[0] = DoesBinaryExist_Prompt($tokens[0]);
    if (not ($tokens[0] eq '')) {
      $gHelper{'more'} = join(' ', @tokens); # This is _already_ a shell string
    }
  }
  if ($gHelper{'more'} eq '') {
    $gHelper{'more'} = DoesBinaryExist_Prompt('more');
    if ($gHelper{'more'} eq '') {
      error('Unable to continue.' . "\n\n");
    }
    $gHelper{'more'} = shell_string($gHelper{'more'}); # Save it as a shell string
  }

  foreach $program ('uname', 'grep', 'ldd', 'mknod', 'insmod', 'modprobe', 'ifconfig') {
    $gHelper{$program} = DoesBinaryExist_Prompt($program);
    if ($gHelper{$program} eq '') {
      error('Unable to continue.' . "\n\n");
    }
  }
}

# Check the validity of an answer whose type is headerdir
# Return a clean answer if valid, or ''
sub check_answer_headerdir {
  my $answer = shift;
  my $source = shift;
  my $pattern = '@@VMWARE@@';
  my $header_version_uts;
  my $header_smp;
  my $header_page_offset;

  $answer = dir_remove_trailing_slashes($answer);

  if (not (-d $answer)) {
    if ($source eq 'user') {
      print wrap('The path "' . $answer . '" is not an existing directory.' . "\n\n", 0);
    }
    return '';
  }

  if (   (not (-d $answer . '/linux'))
      || (not (-d $answer . '/asm'))
      || (not (-d $answer . '/net'))) {
    if ($source eq 'user') {
      print wrap('The path "' . $answer . '" is an existing directory, but it does not contain at least one of these directories "linux", "asm", "net" as expected.' . "\n\n", 0);
    }
    return '';
  }

  #
  # Check that the running kernel matches the set of header files
  #

  if (not (-r $answer . '/linux/version.h')) {
    if ($source eq 'user') {
      print wrap('The path "' . $answer . '" is a kernel header file directory, but it does not contain the file "linux/version.h" as expected. This can happen if the kernel has never been built, or if you have invoked the "make mrproper" command in your kernel directory. In any case, you may want to rebuild your kernel.' . "\n\n", 0);
    }
    return '';
  }
  $header_version_uts = direct_command(shell_string($gHelper{'echo'}) . ' ' . shell_string('#include <linux/version.h>' . "\n" . $pattern . ' UTS_RELEASE') . ' | ' . shell_string($gHelper{'gcc'}) . ' ' . shell_string('-I' . $answer) . ' -E - | ' . shell_string($gHelper{'grep'}) . ' ' . shell_string($pattern));
  chop($header_version_uts);
  $header_version_uts =~ s/^$pattern \"([^\"]*)\".*$/$1/;
  if (not ($header_version_uts eq $gSystem{'version_uts'})) {
    if ($source eq 'user') {
      print wrap('The directory of kernel headers (version ' . $header_version_uts . ') does not match your running kernel (version ' . $gSystem{'version_uts'} . '). Consequently, even if the compilation of the module was successful, the module would not load into the running kernel.' . "\n\n", 0);
    }
    return '';
  }

  if (not (-r $answer . '/linux/autoconf.h')) {
    if ($source eq 'user') {
      print wrap('The path "' . $answer . '" is a kernel header file directory, but it does not contain the file "linux/autoconf.h" as expected. This can happen if the kernel has never been built, or if you have invoked the "make mrproper" command in your kernel directory. In any case, you may want to rebuild your kernel.' . "\n\n", 0);
    }
    return '';
  }
  $header_smp = direct_command(shell_string($gHelper{'grep'}) . ' CONFIG_SMP ' . shell_string($answer . '/linux/autoconf.h'));
  if (not ($header_smp eq '')) {
    # We found a valid up/smp information
    $header_smp = ($header_smp =~ /^\#define CONFIG_SMP/) ? 'yes' : 'no';
    if (not (lc($header_smp) eq lc($gSystem{'smp'}))) {
      if ($source eq 'user') {
	print wrap('The kernel defined by this directory of header files is ' . (($header_smp eq 'yes') ? 'multiprocessor' : 'uniprocessor') . ', while your running kernel is ' . (($gSystem{'version_uts'} eq 'yes') ? 'multiprocessor' : 'uniprocessor') . '.' . "\n\n", 0);
      }
      return '';
    }
  }

  if (not (-r $answer . '/asm/page.h')) {
    if ($source eq 'user') {
      print wrap('The path "' . $answer . '" is a kernel header file directory, but it does not contain the file "asm/page.h" as expected.' . "\n\n", 0);
    }
    return '';
  }
  $header_page_offset = direct_command(shell_string($gHelper{'echo'}) . ' ' . shell_string('#define __KERNEL__' . "\n" . '#include <asm/page.h>' . "\n" . $pattern . ' __PAGE_OFFSET') . ' | ' . shell_string($gHelper{'gcc'}) . ' ' . shell_string('-I' . $answer) . ' -E - | ' . shell_string($gHelper{'grep'}) . ' ' . shell_string($pattern));
  chop($header_page_offset);
  $header_page_offset =~ s/^$pattern \(0x([0-9a-fA-F]{8}).*$/$1/;
  if ($header_page_offset =~ /[0-9a-fA-F]{8}/) {
    # We found a valid page offset
    if (not (lc($header_page_offset) eq lc($gSystem{'page_offset'}))) {
      if ($source eq 'user') {
	print wrap('The kernel defined by this directory of header files does not have the same address space size as your running kernel.' . "\n\n", 0);
      }
      return '';
    }
  }

  return $answer;
}
$gAnswerSize{'headerdir'} = 20;
$gCheckAnswerFct{'headerdir'} = \&check_answer_headerdir;

# Check the validity of an answer whose type is ip
# Return a clean answer if valid, or ''
sub check_answer_ip {
  my $answer = shift;
  my $source = shift;

  # I'm in love with regular expressions --hpreg
  if ($answer =~ /^([0-9]|[1-9][0-9]|1[0-9][0-9]|2([0-4][0-9]|5[0-5]))(\.([0-9]|[1-9][0-9]|1[0-9][0-9]|2([0-4][0-9]|5[0-5]))){3}$/) {
    return $answer;
  }

  if ($source eq 'user') {
    print wrap('The answer "' . $answer . '" is invalid. It must be of the form a.b.c.d where a, b, c and d are decimal numbers between 0 and 255.' . "\n\n", 0);
  }
  return '';
}
$gAnswerSize{'ip'} = 15;
$gCheckAnswerFct{'ip'} = \&check_answer_ip;

# Check the validity of an answer whose type is yesnohelp
# Return a clean answer if valid, or ''
sub check_answer_yesnohelp {
  my $answer = shift;
  my $source = shift;

  if (lc($answer) =~ /^y(es)?$/) {
    return 'yes';
  }

  if (lc($answer) =~ /^n(o)?$/) {
    return 'no';
  }

  if (lc($answer) =~ /^h(elp)?$/) {
    return 'help';
  }

  if ($source eq 'user') {
    print wrap('The answer "' . $answer . '" is invalid. It must be one of "y", "n" or "h".' . "\n\n", 0);
  }
  return '';
}
$gAnswerSize{'yesnohelp'} = 4;
$gCheckAnswerFct{'yesnohelp'} = \&check_answer_yesnohelp;

# Check the validity of an answer whose type is availethif
# Return a clean answer if valid, or ''
sub check_answer_availethif {
  my $answer = shift;
  my $source = shift;

  if (grep(/$answer/, @gAvailEthIf)) {
    return $answer;
  }

  if ($source eq 'user') {
    print wrap('The answer "' . $answer . '" is invalid. It must be one of ' . join(', ', @gAvailEthIf) . '.' . "\n\n", 0);
  }
  return '';
}
$gAnswerSize{'availethif'} = 4;
$gCheckAnswerFct{'availethif'} = \&check_answer_availethif;

# Display the end-user license agreement
sub show_EULA {
  if (   (not defined($gDBAnswer{'EULA_AGREED'}))
      || (db_get_answer('EULA_AGREED') eq 'no')) {
    query('You must read and accept the End User License Agreement to continue.' . "\n" . 'Press enter to display it.', '', 0);

    # $gHelper{'more'} is already a shell string
    system($gHelper{'more'} . ' ' . shell_string(db_get_answer('DOCDIR') . '/EULA'));
    print "\n";

    # Make sure there is no default answer here
    if (get_persistent_answer('Do you accept? (yes/no)', 'EULA_AGREED', 'yesno', '') eq 'no') {
      print wrap('Please try again when you are ready to accept.' . "\n\n", 0);
      exit 0;
    }

    print wrap('Thank you.' . "\n\n", 0);
  }
}

# Build a Linux kernel integer version
sub kernel_version_integer {
  my $version = shift;
  my $patchLevel = shift;
  my $subLevel = shift;

  return $version * 65536 + $patchLevel * 256 + $subLevel;
}

# Retrieve distribution information
sub distribution_info {
  my $issue = '/etc/issue';
  my $system;

  # First use the accurate method that are intended to work reliably on recent
  # distributions (if an FHS guy is listening, we really need a generic way to
  # do this)
  if (-e '/etc/debian_version') {
    return 'debian';
  }
  if (-e '/etc/redhat-release') {
    return 'redhat';
  }
  if (-e '/etc/SuSE-release') {
    return 'suse';
  }
  if (-e '/etc/turbolinux-release') {
    return 'turbolinux';
  }

  # Then use less accurate methods that should work even on old distributions,
  # if people haven't customized their system too much
  if (-e $issue) {
    if (not (direct_command(shell_string($gHelper{'grep'}) . ' -i ' . shell_string('debian') . ' ' . shell_string($issue)) eq '')) {
      return 'debian';
    }
    if (not (direct_command(shell_string($gHelper{'grep'}) . ' -i ' . shell_string('red *hat') . ' ' . shell_string($issue)) eq '')) {
      return 'redhat';
    }
    if (not (direct_command(shell_string($gHelper{'grep'}) . ' -i ' . shell_string('suse\|s\.u\.s\.e') . ' ' . shell_string($issue)) eq '')) {
      return 'suse';
    }
    if (not (direct_command(shell_string($gHelper{'grep'}) . ' -i ' . shell_string('caldera') . ' ' . shell_string($issue)) eq '')) {
      return 'caldera';
    }
  }

  return 'unknown';
}

# Retrieve and check system information
sub system_info {
  my $fullVersion;
  my $version;
  my $patchLevel;
  my $subLevel;
  my @fields;

  $gSystem{'system'} = direct_command(shell_string($gHelper{'uname'}) . ' -s');
  chop($gSystem{'system'});
  if (not ($gSystem{'system'} eq 'Linux')) {
    error('You are not running Linux. This version of the product only runs on Linux.' . "\n\n");
  }

  $gSystem{'version_uts'} = direct_command(shell_string($gHelper{'uname'}) . ' -r');
  chop($gSystem{'version_uts'});

  ($version, $patchLevel, $subLevel) = split(/\./, $gSystem{'version_uts'});
  # Clean the subLevel in case there is an extraversion
  ($subLevel) = split(/[^0-9]/, $subLevel);
  $gSystem{'version_utsclean'} = $version . '.' . $patchLevel . '.' . $subLevel;

  $gSystem{'version_integer'} = kernel_version_integer($version, $patchLevel, $subLevel);
  if ($gSystem{'version_integer'} < kernel_version_integer(2, 0, 0)) {
    error('You are running Linux version ' . $gSystem{'version_utsclean'} . '. This product only runs on 2.0.0 and later kernels.' . "\n\n");
  }

  # CONFIG_UMISC on 2.0 kernels
  if ($gSystem{'version_integer'} < kernel_version_integer(2, 1, 0)) {
    if ((direct_command(shell_string($gHelper{'grep'}) . ' ' . shell_string('^[0-9a-f]* misc_register') . ' /proc/ksyms') eq '') || (direct_command(shell_string($gHelper{'grep'}) . ' ' . shell_string('^[0-9a-fA-F]\{8\} misc_deregister') . ' /proc/ksyms') eq '')) {
      error('You are running a Linux kernel version ' . $gSystem{'version_utsclean'} . ' that was not built with the CONFIG_UMISC configuration parameter set. VMware will not run on this system.' . "\n\n");
    }
  }

  $gSystem{'smp'} = (direct_command(shell_string($gHelper{'uname'}) . ' -v') =~ / SMP /) ? 'yes' : 'no';

  $gSystem{'distribution'} = distribution_info();

  $gSystem{'page_offset'} = 'C0000000';
  @fields = split(' ', direct_command(shell_string($gHelper{'grep'}) . ' ' . shell_string('^[0-9a-fA-F]\{8\} printk') . ' /proc/ksyms'));
  if (defined($fields[0])) {
    my $first;

    $first = lc(substr($fields[0], 0, 1));
    if ($first =~ /^[01234567]$/) {
      # old kernels without the hardware verify_area() support, but that's ok
      # because those kernel didn't have big memory support either.
      $first = 'C';
    } elsif ($first =~ /^[89ab]$/) {
      $first = '8';
    } elsif ($first =~ /^[cdef]$/) {
      $first = 'C';
    }
    $gSystem{'page_offset'} = $first . '0000000';
  }

  # 3Com bug on 2.0.3[45] kernels
  if (   ($gSystem{'version_integer'} >= kernel_version_integer(2, 0, 34))
      && ($gSystem{'version_integer'} <= kernel_version_integer(2, 0, 35))) {
    if (   (not (-r '/proc/ioports'))
	|| (not (direct_command(shell_string($gHelper{'grep'}) . ' -i ' . shell_string('3c90\|3c59') . ' /proc/ioports') eq ''))) {
      if (get_answer('You are running Linux version ' . $gSystem{'version_utsclean'} . ' possibly with a 3Com networking card. Linux kernel versions 2.0.34 and 2.0.35 have a bug in the 3Com driver that interacts badly with this product. Specifically, your physical machine will occasionally hang and will require a hard reset. This bug has been fixed in 2.0.36 and later kernels. Do you want to continue the configuration anyway?', 'yesno', 'no') eq 'no') {
	exit 1;
      }
    }
  }

  # C library
  # XXX This relies on the locale
  if (system(shell_string($gHelper{'ldd'}) . ' ' . shell_string(db_get_answer('BINDIR') . '/vmware') . ' | ' . shell_string($gHelper{'grep'}) . ' -q -i ' . shell_string('not found')) == 0) {
    print wrap('The correct version of one or more libraries needed to run vmware may be missing. This is the output of ' . $gHelper{'ldd'} . ' ' . db_get_answer('BINDIR') . '/vmware:' . "\n", 0);
    system(shell_string($gHelper{'ldd'}) . ' ' . shell_string(db_get_answer('BINDIR') . '/vmware'));
    print "\n";
    query('This script cannot tell for sure, but you may need to upgrade libc5 to glibc before you can run vmware.' . "\n\n" . 'Hit enter to continue.', '', 0);
  }

  # Processor
  if (direct_command(shell_string($gHelper{'grep'}) . ' ' . shell_string('^cpuid') . ' /proc/cpuinfo') eq '') {
    error('Your ' . (($gSystem{'smp'} eq 'yes') ? 'processors do' : 'processor does') . ' not support the cpuid instruction. VMware will not run on this system.' . "\n\n");
  }
  if (direct_command(shell_string($gHelper{'grep'}) . ' ' . shell_string('^flags.* tsc') . ' /proc/cpuinfo') eq '') {
    error('Your ' . (($gSystem{'smp'} eq 'yes') ? 'processors do' : 'processor does') . ' not have a Time Stamp Counter. VMware will not run on this system.' . "\n\n");
  }
}

# Install a module if it suitable
# Return 1 if success, 0 if failure
sub try_module {
  my $name = shift;
  my $mod = shift;
  my $force = shift;
  my $dst_dir;
  my %patch;

  if (not (-e $mod)) {
    # The module does not exist
    return 0;
  }

  # Probe the module without loading it or executing its code. It is cool
  # because it avoids problems like 'Device or resource busy'
  # Note: -f bypasses only the kernel version check, not the symbol resolution
  if (system(shell_string($gHelper{'insmod'}) . ' -p ' . ($force ? '-f ' : '') . shell_string($mod) . ' >/dev/null 2>&1')) {
    return 0;
  }
  
  if (-d $cKernelModuleDir . '/preferred') {
    $dst_dir = $cKernelModuleDir . '/preferred';
  } elsif (-d $cKernelModuleDir . '/'. $gSystem{'version_uts'}) {
    $dst_dir = $cKernelModuleDir . '/' . $gSystem{'version_uts'};
  } else {
    error('Unable to find the directory containing the modules for the running kernel.' . "\n\n");
  }
  create_dir($dst_dir . '/misc', 1);
  undef %patch;
  # Install the module with a .o extension, as the Linux kernel does
  install_file($mod, $dst_dir . '/misc/' . $name . '.o', \%patch, 1);
  # The old installer allowed people to manually build modules without .o
  # extension. Such modules were not removed by the old uninstaller, and
  # unfortunately, insmod tries them first. Let's move them.
  if (file_name_exist($dst_dir . '/misc/' . $name)) {
    backup_file($dst_dir . '/misc/' . $name);
    if (not unlink($dst_dir . '/misc/' . $name)) {
      print STDERR wrap('Unable to remove the file ' . $dst_dir . '/misc/' . $name . '.' . "\n\n", 0);
    }
  }

  return 1;
}

# Remove the build directory
sub remove_build_dir {
  my $dir = shift;

  if (system(shell_string($gHelper{'rm'}) . ' -rf ' . shell_string($dir))) {
    error('Unable to remove the temporary build directory ' . $dir . '.' . "\n\n");
  };
}

# Build a module
sub build_module {
  my $name = shift;
  my $dir = shift;
  my $ideal = shift;
  my $build_dir;

  # Lazy initialization
  if (not defined($gHelper{'make'})) {
    my $program;

    foreach $program ('make', 'gcc', 'echo', 'tar', 'rm') {
      $gHelper{$program} = DoesBinaryExist_Prompt($program);
      if ($gHelper{$program} eq '') {
	error('Unable to continue.' . "\n\n");
      }
    }

    get_persistent_answer('What is the location of the directory of C header files that match your running kernel?', 'HEADER_DIR', 'headerdir', '/usr/src/linux/include');
  }

  print wrap('Extracting the sources of the ' . $name . ' module.' . "\n\n", 0);
  # Do the work on a local filsystem to avoid NFS vs. root permission problems.
  $build_dir = defined($ENV{TMPDIR}) ? $ENV{TMPDIR} : '/tmp';
  $build_dir .= '/vmware/config/' . $$;
  remove_build_dir($build_dir);
  create_dir($build_dir, 0);

  if (system(shell_string($gHelper{'tar'}) . ' -C ' . shell_string($build_dir) . ' -xopf ' . shell_string($dir . '/' . $name . '.tar'))) {
    error('Unable to untar the file ' . $dir . '/' . $name . '.tar' . ' in the ' . $build_dir . 'directory.' . "\n\n");
  }

  print wrap('Building the ' . $name . ' module.' . "\n\n", 0);
  if (system(shell_string($gHelper{'make'}) . ' -C ' . shell_string($build_dir . '/' . $name . '-only') . ' auto-build ' . (($gSystem{'smp'} eq 'yes') ? 'SUPPORT_SMP=1 ' : ' ') . shell_string('HEADER_DIR=' . db_get_answer('HEADER_DIR')))) {
    error('Unable to build the ' . $name . ' module.' . "\n\n");
  }

  # Don't use the force flag: the module is supposed to perfectly load
  if (try_module($name, $build_dir . '/' . $name . '.o', 0)) {
    print wrap('The module has perfectly been loaded in the running kernel.' . "\n\n", 0);
    remove_build_dir($build_dir);
    return;
  }

  # Don't remove the build dir so that the user can investiguate
  error('Unable to make a ' . $name . ' module that can be loaded in the running kernel. There is probably a light difference of kernel configuration between the set of C header files you specified and your running kernel. You may want to rebuild a kernel based on that directory, or specify another directory.' . "\n\n");
}

# Create a list of modules suitable for the running kernel
# The kernel module loader does quite a good job when modules are versioned.
# But in the other case, we must be _very_ careful
sub get_suitable_modules {
  my $dir = shift;
  my @list;
  my $candidate;

  @list = ();
  foreach $candidate (internal_ls($dir)) {
    my %prop;

    # Read the properties file
    if (not open(PROP, '<' . $dir . '/' . $candidate . '/properties')) {
      print STDERR wrap('Unable to open the property file "' . $dir . '/' . $candidate . '/properties". Skipping this kernel.' . "\n\n");
      next;
    }
    undef %prop;
    while (<PROP>) {
      if (/^(\S+) (\S+)/) {
	$prop{$1} = $2;
      }
    }
    close(PROP);

    if (not (lc($gSystem{'smp'}) eq lc($prop{'SMP'}))) {
      # SMP does not match
      next;
    }
    if (not (lc($gSystem{'page_offset'}) eq lc($prop{'PageOffset'}))) {
      # Page offset does not match
      next;
    }

    if ($gSystem{'version_uts'} eq $prop{'UtsRelease'}) {
      # Perfect match. Try this module first
      unshift(@list, ($candidate, $prop{'ModVersion'}));
    } else {
      push(@list, ($candidate, $prop{'ModVersion'}));
    }
  }

  return @list;
}

# Configure a module
sub configure_module {
  my $name = shift;
  my $mod_dir;
  my @mod_list;

  $mod_dir = db_get_answer('LIBDIR') . '/modules';
  @mod_list = get_suitable_modules($mod_dir . '/binary');
  while ($#mod_list > -1) {
    my $candidate = shift(@mod_list);
    my $modversion = shift(@mod_list);

    if (try_module($name, $mod_dir . '/binary/' . $candidate . '/objects/' . $name . '.o', $modversion eq 'yes')) {
      return;
    }
  }

  if (get_persistent_answer('None of VMware' . "'" . 's pre-built ' . $name . ' modules is suitable for your running kernel. Do you want this script to try to build the ' . $name . ' module for your system (you need to have a C compiler installed on your system)?', 'BUILDR_' . $name, 'yesno', 'yes') eq 'no') {
    error('Unable to continue.' . "\n\n");
  }
  build_module($name, $mod_dir . '/source');
}

# Create a character device
sub configure_chrdev {
  my $name = shift;
  my $major = shift;
  my $minor = shift;

  uninstall_file($name);
  if (-e $name) {
    if (-c $name) {
      my @statbuf;

      @statbuf = stat($name);
      if (   defined($statbuf[6])
          && (($statbuf[6] >> 8) == $major)
          && (($statbuf[6] & 0xFF) == $minor)) {
	# The character device is already correctly configured
	return;
      }
    }

    if (get_answer('This script wanted to create the character device ' . $name . ' with major number ' . $major . ' and minor number ' . $minor . ', but there is already a different kind of file at this location. Overwrite?', 'yesno', 'yes') eq 'no') {
      error('Unable to continue.' . "\n\n");
    }

    # mknod doesn't like when the file already exists
    unlink($name);
  }
  if (system(shell_string($gHelper{'mknod'}) . ' ' . shell_string($name) . ' c ' . shell_string($major) . ' ' . shell_string($minor))) {
    error('Unable to create the character device ' . $name . ' with major number ' . $major . ' and minor number ' . $minor . '.' . "\n\n");
  }
  safe_chmod(0600, $name);
  # These file don't have a content, don't timestamp them
  db_add_file($name, 0);
}

# Configuration related to the monitor
sub configure_mon {
  configure_module('vmmon');
  configure_chrdev('/dev/vmmon', 10, 165);
}

# Configuration related to parallel ports
sub configure_pp {
  my $i;

  if ($gSystem{'version_integer'} < kernel_version_integer(2, 1, 127)) {
    query('You are running Linux version ' . $gSystem{'version_utsclean'} . ', and this kernel can not provide VMware with Bidirectional Parallel Port support. A fully-featured VMware requires Linux version 2.1.127 or higher.' . "\n\n" . 'Without this support, VMware will run flawlessly, but will lack the ability to use parallel ports in a bidirectional way. This means that it is possible that some parallel port devices (scanners, dongles, ...) will not work inside a Virtual Machine.' . "\n\n" . 'Hit enter to continue.', '', 0);
    return;
  }

  if ($gSystem{'version_integer'} > kernel_version_integer(2, 3, 9)) {
    query('You are running Linux version ' . $gSystem{'version_utsclean'} . ', and VMware does not provide support for Bidirectional Parallel Ports for Linux version 2.3.10 or higher yet.' . "\n\n" . 'Without this support, VMware will run flawlessly, but will lack the ability to use parallel ports in a bidirectional way. This means that it is possible that some parallel port devices (scanners, dongles, ...) will not work inside a Virtual Machine.' . "\n\n" . 'Hit enter to continue.', '', 0);
    return;
  }

  # The vmppuser module relies on the parport modules. Let's
  # make sure it is loaded before beginning our tests
  if (direct_command(shell_string($gHelper{'grep'}) . '  ' . shell_string(' parport_release[^' . "\t" . ']*$') . ' /proc/ksyms') eq '') { 
    # This comment fixes emacs's broken syntax highlighting
    # parport support is not built in the kernel
    if (system(shell_string($gHelper{'modprobe'}) . ' parport')) {
      query('Unable to load the parport module that is required by the vmppuser module. You may want to load it manually before re-running this script.' . "\n\n" . 'Without this support, VMware will run flawlessly, but will lack the ability to use parallel ports in a bidirectional way. This means that it is possible that some parallel port devices (scanners, dongles, ...) will not work inside a Virtual Machine.' . "\n\n" . 'Hit enter to continue.', '', 0);
      return;
    }
  }

  # The vmppuser module relies on the parport_pc modules. Let's
  # make sure it is loaded before beginning our tests
  if (direct_command(shell_string($gHelper{'grep'}) . '  ' . shell_string(' parport_pc_[^' . "\t" . ']*$') . ' /proc/ksyms') eq '') { 
    # This comment fixes emacs's broken syntax highlighting
    # parport_pc support is not built in the kernel
    if (system(shell_string($gHelper{'modprobe'}) . ' parport_pc')) {
      query('Unable to load the parport_pc module that is required by the vmppuser module. You may want to load it manually before re-running this script.' . "\n\n" . 'Without this support, VMware will run flawlessly, but will lack the ability to use parallel ports in a bidirectional way. This means that it is possible that some parallel port devices (scanners, dongles, ...) will not work inside a Virtual Machine.' . "\n\n" . 'Hit enter to continue.', '', 0);
      return;
    }
  }

  configure_module('vmppuser');

  # Try to unload the modules. Failure is allowed because some other
  # process could be using them.
  system(shell_string($gHelper{'modprobe'}) . ' -r parport_pc >/dev/null 2>&1');
  system(shell_string($gHelper{'modprobe'}) . ' -r parport >/dev/null 2>&1');

  for ($i = 0; $i < 4; $i++) {
    configure_chrdev('/dev/parport' . $i, 99, $i);
  }
}

# Configuration of bridged networking
sub configure_bridged_net {
  # Get the list of available ethernet interfaces
  # The -a is important because it lists all interfaces (not only those
  # which are up). The vmnet driver knows how to deal with down interfaces.
  open(IFCONFIG, shell_string($gHelper{'ifconfig'}) . ' -a |');
  @gAvailEthIf = ();
  while (<IFCONFIG>) {
    if (/^eth/) {
      my @fields;

      @fields = split(/[ ]+/);
      push(@gAvailEthIf, $fields[0]);
    }
  }

  if ($#gAvailEthIf == -1) {
    # No interface. We provide a valid default so that everything works.
    db_add_answer('VNET_INTERFACE', 'eth0');
    return;
  }

  if ($#gAvailEthIf == 0) {
    # Only one interface. Use it.
    db_add_answer('VNET_INTERFACE', $gAvailEthIf[0]);
    return;
  }

  # Several interfaces
  get_persistent_answer('Your computer has multiple ethernet network interfaces: ' . join(', ', @gAvailEthIf) . '. Which one do you want the Virtual Machines to use?', 'VNET_INTERFACE', 'availethif', 'eth0');
}

# Probe for an unused private subnet
# Return value is 1 if success, 0 if failure
sub hostonly_probe {
  my $i;
  my @subnets;
  my $tries;
  my $maxTries = 100;
  my $pings;
  my $maxPings = 10;
  # XXX We only consider class C subnets for the moment
  my $netmask = '255.255.255.0';

  # Generate the table of private class C subnets
  @subnets = ();
  for ($i = 0; $i < 255; $i++) {
    $subnets[2 * $i    ] = '192.168.' . $i;
    $subnets[2 * $i + 1] = '172.16.'  . $i;
  }

  print wrap('Probing for an unused private subnet (this can take some time).' . "\n\n", 0);
  $tries = 0;
  $pings = 0;
  srand(time);
  # Beware, 'last' doesn't seem to work in 'do'-'while' loops
  for (;;) {
    my $r;
    my $subnet;
    my $status;

    $tries++;

    $r = int(rand(2 * 256));
    if ($subnets[$r] eq '') {
      # Already tried
      next;
    }
    $subnet = $subnets[$r];
    $subnets[$r] = '';

    # Our convention is that the host OS IP address is <subnet>.1
    $status = system(shell_string(db_get_answer('BINDIR') . '/vmware-ping') . ' -q ' . shell_string($subnet . '.1')) >> 8;
    if ($status == 1) {
      print wrap('The subnet ' . $subnet . '.0/' . $netmask . ' appears to be unused.' . "\n\n", 0);
      db_add_answer('VNET_HOSTONLY_HOSTADDR', $subnet . '.1');
      db_add_answer('VNET_HOSTONLY_NETMASK', $netmask);
      return 1;
    }

    if ($status == 2) {
      last;
    }

    $pings++;
    if (($pings == $maxPings) || ($tries == $maxTries)) {
      last;
    }
  }

  print STDERR wrap('We were unable to locate an unused Class C subnet in the range of private network numbers. For each subnet that we tried we received a response to our ICMP ping packets from a host at the network address intended for assignment to this host machine. Because no private subnet appears to be unused you will need to explicitly specify a network number.' . "\n\n", 0);
  return 0;
}

# Display the DHCP copyright information
sub show_ISC {
  if (not defined($gDBAnswer{'ISC_COPYRIGHT_SEEN'})) {
    query('Press enter to display the DHCP server copyright information.', '', 0);

    # $gHelper{'more'} is already a shell string
    system($gHelper{'more'} . ' ' . shell_string(db_get_answer('DOCDIR') . '/DHCP-COPYRIGHT'));
    print "\n";

    db_add_answer('ISC_COPYRIGHT_SEEN', 'yes');
  }
}

# Compute the subnet associated to a couple IP/netmask
sub compute_subnet {
  my $ip = shift;
  my $netmask = shift;
  my $prefix;

  # XXX This should be computed based on the netmask instead of assuming a
  # class C in case the user has submitted these values
  ($prefix = $ip) =~ s/\.[0-9]+$//;
  return $prefix . '.0';
}

# Compute the broadcast address associated to a couple IP/netmask
sub compute_broadcast {
  my $ip = shift;
  my $netmask = shift;
  my $prefix;

  # XXX This should be computed based on the netmask instead of assuming a
  # class C in case the user has submitted these values
  ($prefix = $ip) =~ s/\.[0-9]+$//;
  return $prefix . '.255';
}

# Write VMware's DHCPd configuration files
sub write_dhcpd_config {
  my %patch;
  my $prefix;

  undef %patch;
  $patch{'%vmnet%'} = $gHostOnlyEthIf;
  $patch{'%hostaddr%'} = db_get_answer('VNET_HOSTONLY_HOSTADDR');
  $patch{'%netmask%'} = db_get_answer('VNET_HOSTONLY_NETMASK');
  $patch{'%network%'} = compute_subnet($patch{'%hostaddr%'}, $patch{'%netmask%'});
  $patch{'%broadcast%'} = compute_broadcast($patch{'%hostaddr%'}, $patch{'%netmask%'});
  # XXX This should be computed based on the netmask instead of assuming a
  # class C in case the user has submitted these values
  ($prefix = $patch{'%hostaddr%'}) =~ s/\.[0-9]+$//;
  $patch{'%range_low%'} = $prefix . '.128';
  $patch{'%range_high%'} = $prefix . '.254';
  install_file(db_get_answer('LIBDIR') . '/configurator/vmnet-dhcpd.conf',  $cInstallDBDir . '/' . $gHostOnlyEthIf . '.conf', \%patch, 1);

  # Create empty lease files
  # They will be modified by vmnet-dhcpd, don't timestamp them
  undef %patch;
  install_file('/dev/null', $cInstallDBDir . '/' . $gHostOnlyEthIf . '.leases', \%patch, 0);
  safe_chmod(0644, $cInstallDBDir . '/' . $gHostOnlyEthIf . '.leases');
  undef %patch;
  install_file('/dev/null', $cInstallDBDir . '/' . $gHostOnlyEthIf . '.leases~', \%patch, 0);
  safe_chmod(0644, $cInstallDBDir . '/' . $gHostOnlyEthIf . '.leases~');
}

# Check the normal dhcp configuration and give advises
sub dhcpd_consultant {
  my $conf;
  my $network;
  my $netmask;

  if (-r '/etc/dhcpd.conf') {
    $conf = '/etc/dhcpd.conf';
  } else {
    return;
  }

  $netmask = db_get_answer('VNET_HOSTONLY_NETMASK');
  $network = compute_subnet(db_get_answer('VNET_HOSTONLY_HOSTADDR'), $netmask);

  # The host has a normal dhcpd setup
  if (direct_command(shell_string($gHelper{'grep'}) . ' ' . shell_string('^[ ' . "\t" . ']*subnet[ ' . "\t" . ']*' . $network) . ' ' . shell_string($conf)) eq '') {
    query('This system appears to have a DHCP server configured for normal use. Beware that you should teach it how not to interfere with VMware' . "'" . 's DHCP server. There are two ways to do this:' . "\n\n" . '1) Modify the file ' . $conf . ' to add something like:' . "\n\n" . 'subnet ' . $network . ' netmask ' . $netmask . ' {' . "\n" . '    # Note: No range is given, vmnet-dhcpd will deal with this subnet.' . "\n" . '}' . "\n\n" . '2) Start your DHCP server with an explicit list of network interfaces to deal with (leaving out ' . $gHostOnlyEthIf . '). e.g.:' . "\n\n" . 'dhcpd eth0' . "\n\n" . 'Consult the dhcpd(8) and dhcpd.conf(5) manual pages for details.' . "\n\n" . 'Hit enter to continue.', '', 0);
  }
}

# Generate an interfaces specification for a samba configuration file
sub samba_make_interfaces {
  my $if;
  my $result;
  my $sep;
  
  # We assume that ifconfig without any command option only display interfaces
  # that are up.
  open(IFCONFIG, shell_string($gHelper{'ifconfig'}) . ' |');
  # XXX I did my best, but this is probably still locale-dependant --hpreg
  $result = '';
  $sep = '';
  while (<IFCONFIG>) {
    if (/^[a-zA-Z]/) {
      my @fields;

      @fields = split(/[ ]+/);
      $if = $fields[0];
    } elsif (/[iI]net/) {
      if (/^[ ]+.*:(\S+)[ ]+.*:(\S+)[ ]+.*:(\S+)/) {
        if (not ($if eq 'lo')) {
          $result .= $sep . $1 . '/' . $3;
	  $sep = ' ';
	}
      }
    }
  }

  return $result;
}

# Check the samba configuration and give advises
sub samba_consultant {
  my $conf;
  my $prefix;
  my $netmask;
  my $hostaddr;

  if (-r '/etc/smb.conf') {
    $conf = '/etc/smb.conf';
  } elsif (-r '/etc/samba/smb.conf') {
    $conf = '/etc/samba/smb.conf';
  } else {
    return;
  }

  $hostaddr = db_get_answer('VNET_HOSTONLY_HOSTADDR');
  $netmask = db_get_answer('VNET_HOSTONLY_NETMASK');
  # XXX This should be computed based on the netmask instead of assuming a
  # class C in case the user has submitted these values
  ($prefix = $hostaddr) =~ s/\.[0-9]+$//;

  # The host has a samba setup
  # XXX $prefix should be grep-escaped (it contains dots...)
  if (direct_command(shell_string($gHelper{'grep'}) . ' ' . shell_string('^[ ' . "\t" . ']*interfaces[ ' . "\t" . ']*=.*' . $prefix) . ' ' . shell_string($conf)) eq '') {
    query('This system appears to have a CIFS/SMB server (Samba) configured for normal use. Note that if you want to offer service to Virtual Machines running on the host-only network, you must modify your ' . $conf . ' file to list the networks Samba should deal with. You can do this by adding a line looking like this one:' . "\n\n" . 'interfaces = ' . samba_make_interfaces() . ' ' . $hostaddr . '/' . $netmask . "\n\n" . 'You may also need to update any related security controls you might have setup such as the "hosts allow" specification.' . "\n\n" . 'Consult the smb.conf(5) manual page for more details.' . "\n\n" . 'Hit enter to continue.', '', 0);
  }
}

# Configuration of hostonly networking
sub configure_hostonly_net {
  my $keep_settings;

  if (get_persistent_answer('Do you want to be able to use host-only networking in your Virtual Machines?', 'VNET_HOSTONLY', 'yesno', 'yes') eq 'no') {
    return;
  }

  $keep_settings = 'no';
  if (   defined($gDBAnswer{'VNET_HOSTONLY_HOSTADDR'})
      && defined($gDBAnswer{'VNET_HOSTONLY_NETMASK'})) {
    $keep_settings = get_answer('Host-only networking is currently configured to use the private subnet ' . compute_subnet($gDBAnswer{'VNET_HOSTONLY_HOSTADDR'}, $gDBAnswer{'VNET_HOSTONLY_NETMASK'}) . '/' . $gDBAnswer{'VNET_HOSTONLY_NETMASK'} . '. Do you want to keep these settings?', 'yesno', 'yes');
  }

  if ($keep_settings eq 'no') {
    for (;;) {
      my $answer;

      $answer = get_answer('Do you want this script to probe for an unused private subnet? (yes/no/help)', 'yesnohelp', 'yes');

      if ($answer eq 'yes') {
	if (hostonly_probe()) {
	  last;
	}
	# Fallback
	$answer = 'no';
      }

      if ($answer eq 'no') {
	get_persistent_answer('What will be the IP address of your host on the private network?', 'VNET_HOSTONLY_HOSTADDR', 'ip', '');
	get_persistent_answer('What will be the netmask of your private network?', 'VNET_HOSTONLY_NETMASK', 'ip', '');
	last;
      }

      print wrap('Virtual machines configured to use host-only networking are placed on a virtual network that is confined to this host. Virtual machines on this network can communicate with each other and the host, but no one else.' . "\n\n" . 'To setup this host-only networking you need to select a network number that is normally unreachable from the host. We can automatically select this number for you, or you can specify a network number that you want.' . "\n\n" . 'The automatic selection process works by testing a series of Class C subnet numbers to see if they are reachable from the host. The first one that is unreachable is used. The subnet numbers are chosen from the private network numbers specified by the Internet Engineering Task Force (IETF) in RFC 1918 (http://www.isi.edu/in-notes/rfc1918.txt).' . "\n\n" . 'Remember that the host-only network that virtual machines reside on will not be accessible outside the host. This means that it is ok to use the same number on different systems so long as you do not enable communication between these networks.' . "\n\n", 0);
    }
  }

  show_ISC();
  write_dhcpd_config();
  dhcpd_consultant();
  samba_consultant();
}

# Configuration related to networking
sub configure_net {
  my $i;

  if (get_persistent_answer('Do you want to be able to use the network in your Virtual Machines?', 'NETWORKING', 'yesno', 'yes') eq 'no') {
    return;
  }

  configure_module('vmnet');
  for ($i = 0; $i < 4; $i++) {
    configure_chrdev('/dev/vmnet' . $i, 119, $i);
  }

  configure_bridged_net();
  configure_hostonly_net();
}

# Install one symbolic link
sub install_symlink {
  my $to = shift;
  my $name = shift;

  uninstall_file($name);
  if (file_check_exist($name)) {
    return;
  }
  # The file could be a symlink to another location. Remove it
  unlink($name);
  if (not symlink($to, $name)) {
    error('Unable to create the symbolic link ' . $name . ' pointing to the file ' . $to . '.' . "\n\n");
  }
  db_add_file($name, 0);
}

# Install a pair of S/K startup scripts for a given runlevel
sub link_runlevel {
   my $level = shift;

   # Create the S symlink
   # We use 90 because samba is at 91 and it didn't like it when we used 99.
   install_symlink(db_get_answer('INITDIR') . '/init.d/vmware', db_get_answer('INITDIR') . '/rc' . $level . '.d/S90vmware');

   # Create the K symlink
   # Note: SuSE 6.0 handles startup scripts differently (I actually like it).
   # When entering a new runlevel, K* scripts of the _old_ runlevel are
   # executed instead of those of the new runlevel
   if (not ($gSystem{'distribution'} eq 'suse')) {
      # Kill the services in the reboot runlevel
      $level = 6;
   }
   install_symlink(db_get_answer('INITDIR') . '/init.d/vmware', db_get_answer('INITDIR') . '/rc' . $level . '.d/K08vmware');
}

# Create the links for VMware's services
sub link_services {
  my @fields;

  @fields = split(/:/, direct_command(shell_string($gHelper{'grep'}) . ' ' . shell_string('^[ ' . "\t" . ']*[^#]*:initdefault') . ' /etc/inittab'));
  if (defined($fields[1]) && ($fields[1] =~ /[0123456Ss]/)) {
    link_runlevel($fields[1]);
  } else {
    # Unable to determine the default runlevel
    link_runlevel(2);
    link_runlevel(3);
    link_runlevel(5);
  }
}

# Write the VMware host-wide configuration file
sub write_vmware_config {
  my $name;
  my $promoconfig;

  $name = $cInstallDBDir . '/config';

  uninstall_file($name);
  if (file_check_exist($name)) {
    return;
  }
  # The file could be a symlink to another location. Remove it
  unlink($name);

  open(CONFIGFILE, '>' . $name) or error('Unable to open the configuration file ' . $name . ' in write-mode.' . "\n\n");
  db_add_file($name, 1);
  safe_chmod(0444, $name);
  print CONFIGFILE 'vmware.fullpath = "' . db_get_answer('BINDIR') . '/vmware"' . "\n";
  print CONFIGFILE 'wizard.fullpath = "' . db_get_answer('BINDIR') . '/vmware-wizard"' . "\n";
  print CONFIGFILE 'dhcpd.fullpath = "' . db_get_answer('BINDIR') . '/vmnet-dhcpd"' . "\n";
  print CONFIGFILE 'loop.fullpath = "' . db_get_answer('BINDIR') . '/vmware-loop"' . "\n";
  print CONFIGFILE 'libdir = "' . db_get_answer('LIBDIR') . '"' . "\n";
  if (defined($gDBAnswer{'NETWORKING'}) && defined($gDBAnswer{'VNET_HOSTONLY_HOSTADDR'})) {
    print CONFIGFILE $gHostOnlyEthIf . '.HostOnlyAddress = "' . db_get_answer('VNET_HOSTONLY_HOSTADDR') . '"' . "\n";
  }
  if (defined($gDBAnswer{'NETWORKING'}) && defined($gDBAnswer{'VNET_HOSTONLY_NETMASK'})) {
    print CONFIGFILE $gHostOnlyEthIf . '.HostOnlyNetMask = "' . db_get_answer('VNET_HOSTONLY_NETMASK') . '"' . "\n";
  }
  close(CONFIGFILE);

  # Append the promotional configuration if it exists
  $promoconfig = db_get_answer('LIBDIR') . '/configurator/PROMOCONFIG';
  if (-e $promoconfig) {
    my %patch;

    undef %patch;
    internal_sed($promoconfig, $name, 1, \%patch);
  }
}

# Display the PROMOCODE information
sub show_PROMOCODE {
  my $promocode;

  $promocode = db_get_answer('DOCDIR') . '/PROMOCODE';
  if (-e $promocode) {
    # $gHelper{'more'} is already a shell string
    system($gHelper{'more'} . ' ' . shell_string($promocode));
    print "\n";
  }
}

# Display a usage error message for the configuration program and exit
sub config_usage {
  print STDERR wrap('VMware ' . vmware_version() . ' for Linux configurator' . "\n" . 'Usage: ' . $0 . ' [[-][-]d[efault]]' . "\n" . '    default: Automatically answer questions with the proposed answer.' . "\n\n", 0);
  exit 1;
}

# Program entry point
sub main {
  if (not is_root()) {
    error('Please re-run this script as the super user.' . "\n\n");
  }

  # Force the path to reduce the risk of using "modified" external helpers
  # If the user has a special system setup, he will will prompted for the
  # proper location anyway
  $ENV{'PATH'} = '/bin:/usr/bin:/sbin:/usr/sbin';

  $gOption{'default'} = 0;
  initialize_external_helpers();

  if ($#ARGV > -1) {
    if ($#ARGV > 0) {
      config_usage();
    }

    # There is only one argument
    if (lc($ARGV[0]) !~ /^(-)?(-)?d(efault)?/) {
      install_usage();
    }

    $gOption{'default'} = 1;
  }

  if (not (-e $cInstallDBFileName)) {
    error('Unable to find the database file (' . $cInstallDBFileName . ')' . "\n\n");
  }
  db_load();

  # Stop VMware's services
  print wrap('Making sure VMware' . "'" . 's services are stopped.' . "\n\n", 0);
  if (system(shell_string(db_get_answer('INITDIR') . '/init.d/vmware') . ' stop')) {
    error('Unable to stop VMware' . "'" . 's services.' . "\n\n");
  }
  print "\n";
  db_append();
  show_EULA();
  system_info();
  configure_mon();
  configure_pp();
  configure_net();
  link_services();
  write_vmware_config();
  # Remove the flag _before_
  uninstall_file($cConfFlag);
  db_save();
  # Then start VMware's services
  system(shell_string(db_get_answer('INITDIR') . '/init.d/vmware') . ' start');
  print "\n";

  show_PROMOCODE();
  print wrap('The configuration of VMware ' . vmware_version() . ' for Linux for this running kernel completed successfully.' . "\n\n" . 'You can now run VMware by invoking the following command: "' . db_get_answer('BINDIR') . '/vmware".' . "\n\n" . 'Enjoy,' . "\n\n" . '    --the VMware team' . "\n\n", 0);
  exit(0);
}

main();
