/* **********************************************************
 * Copyright (C) 1998-1999 VMWare, Inc.
 * All Rights Reserved
 * **********************************************************/
#ifdef VMX86_DEVEL
char rcsId_memtrack[] = "$Id: memtrack.c,v 1.3 1999/06/14 16:19:48 bennett Exp $";
#else
#define FILECODE "F(304)"
#endif 

/*
 * memtrack.c --
 *
 *      Utility module for tracking memory allocated and/or 
 *      locked by the monitor. 
 *     
 *      
 */

#ifdef linux
#   include <linux/string.h> /* Needed for memset() in the kernel */
#else
#   include <ntddk.h>
#endif


#include "hostif.h"
#include "x86.h"
#include "vm_types.h"
#include "vm_assert.h"
#include "memtrack.h"
#include "vmx86.h"


/*
 * MemTrack - We track memory using a two-level page table
 * like structure. We allocate the first level in the granularity 
 * of pages and pack as many MemTrackEntires as possible in a page. 
 * We currently handle 1000*4096/sizeof(Entry) or 341K pages. It seems
 * unlikely we will need that much. 
 */

#define MAX_TRACE_PAGE_DIR 1000

#define HASH_TABLE_SIZE    8192
#define HASH_TABLE_ENTRIES_PER_PAGE (PAGE_SIZE /  (sizeof(void*)))
#define HASH_TABLE_PAGES   (HASH_TABLE_SIZE / HASH_TABLE_ENTRIES_PER_PAGE)


typedef struct MemTrack {
   int sizeOfEntry;          /* Size of an entry */
   int numEntriesPerPage;    /* Number of entry per page */
   int numPages;             /* Number of pages stored in the tracker */
   char *mapPages[MAX_TRACE_PAGE_DIR];
   MemTrackEntry **hashTablePages[HASH_TABLE_PAGES]; /* VPN to entry hashtable */
} MemTrack;


static INLINE MemTrackEntry **
HASH_VPN(MemTrack *memtrack,
         VPN vpn) 
{
   uint32 hash = vpn % HASH_TABLE_SIZE;
   int page = hash / HASH_TABLE_ENTRIES_PER_PAGE;
   int offset = hash % HASH_TABLE_ENTRIES_PER_PAGE;
   return memtrack->hashTablePages[page] + offset;
}


/*
 *----------------------------------------------------------------------
 *
 * MemTrack_Init --
 *
 *      Allocate and initialize the memory tracker.
 *
 * Results:
 *      Handle used to access the memtracker.
 *
 * Side effects:
 *      memory allocation.
 *
 *----------------------------------------------------------------------
 */

void *
MemTrack_Init(int sizeOfEntry)
{
   MemTrack *st;
   int i;
   st = HostIF_AllocKernelMem(sizeof(MemTrack), FALSE);

   if (st == NULL) { 
      Warning("MemTrack_Init failed\n");
      return NULL;
   }
   
   memset(st,0,sizeof(MemTrack));
   
   for (i=0;i<HASH_TABLE_PAGES;i++) { 
      VA page = (VA)HostIF_AllocPage(FALSE);
      if (page == 0) { 
         Warning("MemTrack_Init failed on hashTablPages %d\n",i);
         // LEAK
         return NULL;
      }
      if (page & (PAGE_SIZE-1)) { 
         Warning("MemTrack_Init unaligned 0x%x\n",st->hashTablePages[i]);
         // LEAK
         return NULL;
      }
      st->hashTablePages[i] = (MemTrackEntry **)page;
      memset((char*)page, 0, PAGE_SIZE);
      
   }
   
   ASSERT((sizeOfEntry >= sizeof(MemTrackEntry)) && (sizeOfEntry < PAGE_SIZE));
   st->sizeOfEntry = sizeOfEntry;
   st->numEntriesPerPage = PAGE_SIZE / sizeOfEntry;
   st->numPages = 0;
   memset(st->mapPages, 0, sizeof(st->mapPages));
   return (void *) st;
}

/*
 *----------------------------------------------------------------------
 *
 * MemTrack_Add --
 *
 *      Add the specified address to the memory tracker. 
 *
 * Results:
 *      A pointer to the allocated element, or NULL on error.
 *
 * Side effects:
 *      memory allocation.
 *
 *----------------------------------------------------------------------
 */

MemTrackEntry *
MemTrack_Add(void *s,  // memtracker handler
             VPN vpn,  // VPN of entry
             MPN mpn)  // MPN of entry
{
  MemTrack *st = (MemTrack *) s;
  int ind = st->numPages;
  int dirPage = ind / st->numEntriesPerPage;
  int offset = ind % st->numEntriesPerPage;
  MemTrackEntry *ptr;
  MemTrackEntry **head;
  /*
   * Allocate a directory page if needed.
   */
  if (dirPage >= MAX_TRACE_PAGE_DIR) return NULL;

  if (st->mapPages[dirPage] == NULL) {
     st->mapPages[dirPage] = HostIF_AllocPage(FALSE);
     if (st->mapPages[dirPage] == NULL) return NULL;
  }

  /*
   * Store the page in the tracker 
   */
  ptr = (MemTrackEntry *)((st->mapPages[dirPage] + (offset * st->sizeOfEntry)));
  memset(ptr, 0, st->sizeOfEntry);
  ptr->vpn = vpn;  
  ptr->mpn = mpn;  
  ptr->hashChain = NULL;
  
  head = HASH_VPN(st,vpn);
  if (*head == NULL) {
     *head = ptr;
  } else {
     MemTrackEntry *nextPtr = *head;
     while (nextPtr->hashChain != NULL) {
        nextPtr = nextPtr->hashChain;
     }
     nextPtr->hashChain = ptr;
  }
  st->numPages++;
  
  return ptr;
}

/*
 *----------------------------------------------------------------------
 *
 * MemTrack_LookupVPN --
 *
 *      Lookup the specified address in the memory tracker. 
 *
 * Results:
 *      A pointer to the allocated element, or NULL if not there;
 *
 * Side effects:
 *      memory allocation.
 *
 *----------------------------------------------------------------------
 */

MemTrackEntry *
MemTrack_LookupVPN(void *s,  // memtracker handler
        VPN vpn)  // Value to lookup
{
  MemTrack *st = (MemTrack *) s;
  MemTrackEntry *nextPtr;
    
  nextPtr = *HASH_VPN(st,vpn);
  while (nextPtr != NULL) {
    if (nextPtr->vpn == vpn) {
      return nextPtr;
    }
    nextPtr = nextPtr->hashChain;
  }
  return NULL;
}


/*
 *----------------------------------------------------------------------
 *
 * MemTrack_Scan --
 *
 *      Scan the stored pages using a provide search function. 
 *
 * Results:
 *      The return result of the searchFunc
 *
 * Side effects:
 *      None
 *
 *----------------------------------------------------------------------
 */

void *
MemTrack_Scan(void *s,  // Handle for memtracker
        void *arg,  // Argument to searchFunc
        void *(*searchFunc)(void *arg, MemTrackEntry *ptr))
{
  MemTrack *st = (MemTrack *) s;
  int ind;
  int dirPage;
  void *ptr;
  void *ret;

  /* 
   * Call searchFunc with every address stored in the tracker
   * or until it returns a non-zero value.
   */
  for (ind = 0; ind < st->numPages; ind++) {
    dirPage = ind / st->numEntriesPerPage;
    ptr = st->mapPages[dirPage] + 
      (ind % st->numEntriesPerPage) * st->sizeOfEntry;
    ret = searchFunc(arg, ptr);
    if (ret) return ret;
  }
  return NULL;
}

/*
 *----------------------------------------------------------------------
 *
 * MemTrack_Cleanup --
 *
 *      Cleanup all resources allocated for the memtracker. For
 *      all pages in the tracker call the user provided free function.
 *
 * Results:
 *      Number of pages in the tracker. 
 *
 * Side effects:
 *      Memory free
 *
 *----------------------------------------------------------------------
 */

int 
MemTrack_Cleanup(void *s,   // Mem tracker handle
                 void (*CleanUp)(void *clientData,MemTrackEntry*),
                 void *clientData)   // free functions.
{
  MemTrack *st = (MemTrack *) s;
  int ind;
  int dirPage;
  void * ptr;
  int count = 0;
  int i;

  for (ind = 0; ind < st->numPages; ind++) {
    dirPage = ind / st->numEntriesPerPage;
    ptr = st->mapPages[dirPage] + 
      (ind % st->numEntriesPerPage) * st->sizeOfEntry;
    CleanUp(clientData,ptr);
    count++;
  }
  /*
   * Free any directory page number and then free the tracker
   * structure itself 
   */
  for (dirPage = 0; dirPage < MAX_TRACE_PAGE_DIR; dirPage++) {
     if (st->mapPages[dirPage]) { 
        HostIF_FreePage(st->mapPages[dirPage]);
        st->mapPages[dirPage] = NULL;
     }
  }
  
  for (i=0;i<HASH_TABLE_PAGES;i++) { 
     if (st->hashTablePages[i]) {
        HostIF_FreePage(st->hashTablePages[i]);
        st->hashTablePages[i] = NULL;
     }
  }

  HostIF_FreeKernelMem(st);
  return count;
}

